# Start Notice
# This program is a patent protected modification of Chacha20 related Python program developed
# and posted by Jack O'Connor at https://github.com/oconnor663/pure_python_salsa_chacha.
# The original prorans are in the public domain. This is specifically and expressly NOT the case for the modifications made by us.
# The copyright of the modification as well as its  Intelectual Property (IP) are owned by LCIP JV.
#  LCIP JV provides a license to use the programs herein ONLY for education, testing, demonstration
# and research. Operational application in encrypting/decrypting data for data protection as far the protected IP
# is concerned is strictly prohibited in the jurisdiction of the IP. Additionally, as provided in this form here, the code is set up for demonstration only
# and is not secure. 
# Despite seemingly simple modification its plaintext is difficult to re-engineer if one changes one or more of the parameters and/or
# modifications. We cannot and will not try to recover any plaintext from ciphertext.
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Any copy of the program should be made with inclusion of this notice.
# Contact us at info@labcyfer.com
# End Notice 

import struct
import scipy.io
import numpy as np
import ctypes

def mask32(x):
    return x & 0xFFFFFFFF


def add32(x, y):
    return mask32(x + y)


def left_rotate(x, n):
    return mask32(x << n) | (x >> (32 - n))

def dec2rad(a):
    # Ensure a is a 32-bit integer
    a = int(a) & 0xffffffff
    # Partition the 32-bit word into 4 bytes
    byte1 = (a >> 24) & 0xff
    byte2 = (a >> 16) & 0xff
    byte3 = (a >> 8) & 0xff
    byte4 = a & 0xff
    # Output a vector of 4 256-state elements
    return [byte1, byte2, byte3, byte4]

def rad2dec(ar):
    # Combine the 4 bytes into a single 32-bit word
    #res = (ar[0] << 24) | (ar[1] << 16) | (ar[2] << 8) | ar[3]
    #res = ctypes.c_uint32(res).value
    res = struct.unpack('>I', bytes(ar))[0]
    return res



# a, b, c, and d are indexes into the 16-word block
def quarter_roundflt(block, a, b, c, d, funrev):
    block[a] = add32(block[a], block[b])

    dr=dec2rad(block[d])
    ar=dec2rad(block[a])
    dr=[funrev[ar[i], dr[i]] for i in range(4)]
    block[d]=rad2dec(dr)
   

    #block[d] ^= block[a] 
    block[d] = left_rotate(block[d], 16)
    block[c] = add32(block[c], block[d])

    #block[b] ^= block[c]
    br=dec2rad(block[b])
    cr=dec2rad(block[c])
    bbr=[funrev[br[i], cr[i]] for i in range(4)]
    block[b]=rad2dec(bbr)

    block[b] = left_rotate(block[b], 12)
    block[a] = add32(block[a], block[b])

    #block[d] ^= block[a]
    dr=dec2rad(block[d])
    ar=dec2rad(block[a])
    ddr=[funrev[dr[i], ar[i]] for i in range(4)]
    block[d]=rad2dec(ddr)

    block[d] = left_rotate(block[d], 8)
    block[c] = add32(block[c], block[d])

    #block[b] ^= block[c]
    br=dec2rad(block[b])
    cr=dec2rad(block[c])
    bbr=[funrev[br[i], cr[i]] for i in range(4)]
    block[b]=rad2dec(bbr)

    block[b] = left_rotate(block[b], 7)



# a, b, c, and d are indexes into the 16-word block
def quarter_round(block, a, b, c, d):
    #print('for test block[a]=  :', block[a])
    block[a] = add32(block[a], block[b])
    block[d] ^= block[a]
    block[d] = left_rotate(block[d], 16)
    block[c] = add32(block[c], block[d])
    block[b] ^= block[c]
    block[b] = left_rotate(block[b], 12)
    block[a] = add32(block[a], block[b])
    block[d] ^= block[a]
    block[d] = left_rotate(block[d], 8)
    block[c] = add32(block[c], block[d])
    block[b] ^= block[c]
    block[b] = left_rotate(block[b], 7)


def chacha20_permute(block):
    data_funrev = scipy.io.loadmat('sn256.mat')
    funrev = data_funrev['sn256']
   # print(' ')
   # print('start of encrypt funrev: ', funrev[0:5, 0:5])
   # print(' ')

    for doubleround in range(10):
        if doubleround==5:
          
           quarter_roundflt(block, 0, 4, 8, 12, funrev)  # column 1
           quarter_roundflt(block, 1, 5, 9, 13, funrev)  # column 2
           quarter_roundflt(block, 2, 6, 10, 14, funrev)  # column 3
           quarter_roundflt(block, 3, 7, 11, 15, funrev)  # column 4
           quarter_roundflt(block, 0, 5, 10, 15, funrev)  # diagonal 1
           quarter_roundflt(block, 1, 6, 11, 12, funrev)  # diagonal 2
           quarter_roundflt(block, 2, 7, 8, 13, funrev)  # diagonal 3
           quarter_roundflt(block, 3, 4, 9, 14, funrev)  # diagonal 4
        else:
           
           quarter_round(block, 0, 4, 8, 12)  # column 1
           quarter_round(block, 1, 5, 9, 13)  # column 2
           quarter_round(block, 2, 6, 10, 14)  # column 3
           quarter_round(block, 3, 7, 11, 15)  # column 4
           quarter_round(block, 0, 5, 10, 15)  # diagonal 1
           quarter_round(block, 1, 6, 11, 12)  # diagonal 2
           quarter_round(block, 2, 7, 8, 13)  # diagonal 3
           quarter_round(block, 3, 4, 9, 14)  # diagonal 4


def words_from_bytes(b):
    assert len(b) % 4 == 0
    return [int.from_bytes(b[4 * i : 4 * i + 4], "little") for i in range(len(b) // 4)]


def bytes_from_words(w):
    return b"".join(word.to_bytes(4, "little") for word in w)


# This is the IETF (RFC 7539) version of ChaCha20, with a 96-bit nonce.
def chacha20_block(key, nonce, blocknum):
    # This implementation doesn't support 16-byte keys.
    assert len(key) == 32
    assert len(nonce) == 12
    assert blocknum < 2 ** 32
    constant_words = words_from_bytes(b"expand 32-byte k")
    key_words = words_from_bytes(key)
    nonce_words = words_from_bytes(nonce)
    # fmt: off
    original_block = [
        constant_words[0],  constant_words[1],  constant_words[2],  constant_words[3],
        key_words[0],       key_words[1],       key_words[2],       key_words[3],
        key_words[4],       key_words[5],       key_words[6],       key_words[7],
        mask32(blocknum),   nonce_words[0],     nonce_words[1],     nonce_words[2],
    ]
    # fmt: on
    permuted_block = list(original_block)

    chacha20_permute(permuted_block)
    for i in range(len(permuted_block)):
        permuted_block[i] = add32(permuted_block[i], original_block[i])
    return bytes_from_words(permuted_block)


def chacha20_stream(key, nonce, length):
    output = bytearray()
    blocknum = 0
    while length > 0:
        block = chacha20_block(key, nonce, blocknum)
        take = min(length, len(block))
        output.extend(block[:take])
        length -= take
        blocknum += 1
    return output


def chacha20_xor(key, nonce, message):
    stream = chacha20_stream(key, nonce, len(message))
    return bytes(x ^ y for x, y in zip(message, stream))


def hchacha20(key, input_bytes):
    # This implementation doesn't support 16-byte keys.
    assert len(key) == 32
    assert len(input_bytes) == 16
    constant_words = words_from_bytes(b"expand 32-byte k")
    key_words = words_from_bytes(key)
    input_words = words_from_bytes(input_bytes)
    # fmt: off
    block = [
        constant_words[0],  constant_words[1],  constant_words[2],  constant_words[3],
        key_words[0],       key_words[1],       key_words[2],       key_words[3],
        key_words[4],       key_words[5],       key_words[6],       key_words[7],
        input_words[0],     input_words[1],     input_words[2],     input_words[3],
    ]
    # fmt: on
    chacha20_permute(block)
    outputs = block[0:4] + block[12:16]
    return bytes_from_words(outputs)

def xchacha20_stream(key, nonce, length):
    # This implementation doesn't support 16-byte keys.
    assert len(key) == 32
    assert len(nonce) == 24
    derived_key = hchacha20(key, nonce[:16])
    return chacha20_stream(derived_key, b"\0\0\0\0" + nonce[16:], length)


def xchacha20_xor(key, nonce, message):
    stream = xchacha20_stream(key, nonce, len(message))
    return bytes(x ^ y for x, y in zip(message, stream))
