# Start Notice
# This program is a patent protected modification of Chacha20 related Python program developed
# and posted by Jack O'Connor at https://github.com/oconnor663/pure_python_salsa_chacha.
# The original prorans are in the public domain. This is specifically and expressly NOT the case for the modifications made by us.
# The copyright of the modification as well as its  Intelectual Property (IP) are owned by LCIP JV.
#  LCIP JV provides a license to use the programs herein ONLY for education, testing, demonstration
# and research. Operational application in encrypting/decrypting data for data protection as far the protected IP
# is concerned is strictly prohibited in the jurisdiction of the IP. Additionally, as provided in this form here, the code is set up for demonstration only
# and is not secure. 
# Despite seemingly simple modification its plaintext is difficult to re-engineer if one changes one or more of the parameters and/or
# modifications. We cannot and will not try to recover any plaintext from ciphertext.
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Any copy of the program should be made with inclusion of this notice.
# Contact us at info@labcyfer.com
# End Notice 

import pure_chacha20flt
import pure_chacha20fltsn
def ascii_to_hex(ascii_str):
    return ascii_str.encode('utf-8').hex()

def test_hchacha20_vector_real():
    # This test vector comes from https://datatracker.ietf.org/doc/html/draft-arciszewski-xchacha-03.
    key = bytes.fromhex(
        "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f"
    )
    input_bytes = bytes.fromhex("000000090000004a0000000031415927")
    expected_keystream = bytes.fromhex(
        "82413b4227b27bfed30e42508a877d73a0f9e4d58a74a853c12ec41326d3ecdc"
    )
    found_keystream = pure_chacha20flt.hchacha20(key, input_bytes)
    assert expected_keystream == found_keystream
    print('correct3')

    # Use the plaintext from test_xchacha20_vector
    plaintext = bytes.fromhex(
        "5468652064686f6c65202870726f6e6f756e6365642022646f6c652229206973"
        + "20616c736f206b6e6f776e2061732074686520417369617469632077696c6420"
        + "646f672c2072656420646f672c20616e642077686973746c696e6720646f672e"
        + "2049742069732061626f7574207468652073697a65206f662061204765726d61"
        + "6e20736865706865726420627574206c6f6f6b73206d6f7265206c696b652061"
        + "206c6f6e672d6c656767656420666f782e205468697320686967686c7920656c"
        + "757369766520616e6420736b696c6c6564206a756d70657220697320636c6173"
        + "736966696564207769746820776f6c7665732c20636f796f7465732c206a6163"
        + "6b616c732c20616e6420666f78657320696e20746865207461786f6e6f6d6963"
        + "2066616d696c792043616e696461652e"
    )

  # Ask for ASCII plaintext input
    ascii_plaintext = input("Enter ASCII plaintext: ")
    plaintext = bytes.fromhex(ascii_to_hex(ascii_plaintext))

    # Encrypt the plaintext using a 12-byte nonce
    nonce = bytes.fromhex("000000090000004a00000000")  # 12-byte nonce
    ciphertext = pure_chacha20flt.chacha20_xor(key, nonce, plaintext)
    print('Ciphertext:', ciphertext.hex())
    recovertext=pure_chacha20fltsn.chacha20_xor(key, nonce, ciphertext)
    print('Recovertext:', recovertext.hex())
    print('Recovertext (ASCII):', recovertext.decode('utf-8'))
    pop=plaintext==recovertext
    print('pop  =', pop)

def test_hchacha20_vector():
    # This test vector comes from https://datatracker.ietf.org/doc/html/draft-arciszewski-xchacha-03.
    key = bytes.fromhex(
        "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f"
    )
    input_bytes = bytes.fromhex("000000090000004a0000000031415927")
    expected = bytes.fromhex(
        "82413b4227b27bfed30e42508a877d73a0f9e4d58a74a853c12ec41326d3ecdc"
    )
    found = pure_chacha20flt.hchacha20(key, input_bytes)
    assert expected == found
    print('correct3')

# Call the test functions
print('start of tests')
test_hchacha20_vector_real()

